def confirm_txn(txn_sig, max_retries=20, retry_interval=3):
    retries = 0
    
    while retries < max_retries:
        try:
            
            txn_res = get_transaction(signature=str(txn_sig))
            if txn_res:
                print(txn_res)
                print(f"\n\nhttps://solscan.io/tx/{str(txn_sig)}")
                break
            txn_json = safe_json_loads(txn_res.get('transaction',{}).get('meta',{}))
            error = txn_json.get('err')
            if error is None:
                print("Transaction confirmed... try count:", retries+1)
                return True
            print("Error: Transaction not confirmed. Retrying...")
            if error:
                print("Transaction failed.")
                return False
        except Exception as e:
            print("Awaiting confirmation... try count:", retries+1)
            retries += 1
            time.sleep(retry_interval)
    print("Max retries reached. Transaction confirmation failed.")
    return None


def buildTxn(mint, amount, slippage, token_account_pubkey,sol_in,token_account_instructions,buy=True):
    # Get keys for the transaction, pass the token account's pubkey instead of the AccountMeta object
    keys = getKeys(get_coin_data(mint), token_account_pubkey, owner,buy=buy)
    
    if buy:
        # Calculate max_sol_cost
        slippage_adjustment = 1 + (slippage / 100)
        sol_in_with_slippage = sol_in * slippage_adjustment
        max_sol_cost = int(sol_in_with_slippage * LAMPORTS_PER_SOL)
        print("Max Sol Cost:", sol_in_with_slippage)
        hex_data = bytes.fromhex("66063d1201daebea")
        solCost = max_sol_cost
    else:
        # Calculate minimum SOL output
        sol_out = float(token_balance) * float(token_price)
        slippage_adjustment = 1 - (slippage / 100)
        sol_out_with_slippage = sol_out * slippage_adjustment
        min_sol_output = int(sol_out_with_slippage * LAMPORTS_PER_SOL)
        print("Min Sol Output:", sol_out_with_slippage)
        hex_data = bytes.fromhex("33e685a4017f83ad")
        solCost = min_sol_output
    
    data = bytearray()
    data.extend(hex_data)
    data.extend(struct.pack('<Q', amount))
    data.extend(struct.pack('<Q', solCost))
    data = bytes(data)
    swap_instruction = Instruction(PUMP_FUN_PROGRAM, data, keys)
    blockHash = requests.post(url="https://rpc.ankr.com/solana/c3b7fd92e298d5682b6ef095eaa4e92160989a713f5ee9ac2693b4da8ff5a370",data=json.dumps({"id":1,"jsonrpc":"2.0","method":"getLatestBlockhash","params":[{"commitment":"processed"}]}))
    recent_blockhash = get_any_value(blockHash.json(),'blockhash')
    recent_blockhash = Hash.from_string(recent_blockhash)
    txn = Transaction(recent_blockhash=recent_blockhash, fee_payer=owner)
    txn.add(set_compute_unit_price(UNIT_PRICE))
    txn.add(set_compute_unit_limit(UNIT_BUDGET))
    
    if buy:
        if token_account_instructions:
            txn.add(token_account_instructions)
        txn.add(swap_instruction)
    else:
        txn.add(swap_instruction)
        if close_token_account:
            close_account_instructions = close_account(CloseAccountParams(TOKEN_PROGRAM, token_account_pubkey, owner, owner))
            txn.add(close_account_instructions)
    
    txn.sign(payer_keypair)
    # Send and confirm transaction
    txn_sig = sendTransaction(txn, payer_keypair, opts=TxOpts(skip_preflight=True))
    print("Transaction Signature", txn_sig)
    confirm = confirm_txn(txn_sig)
    print(confirm)



def pump_fun_buy(mint_str: str, sol_in: float = 0.001, slippage: int = 25) -> bool:
    coin_data = get_coin_data(mint_str)
    print("Coin Data:", coin_data)

    if not coin_data:
        print("Failed to retrieve coin data...")
        return False

    payer_keypair = Keypair.from_base58_string(get_env_value(key='AMM_P'))
    owner = payer_keypair.pubkey()
    print("Owner Public Key:", owner)

    mint = Pubkey.from_string(mint_str)
    token_account, token_account_instructions = check_existing_token_account(owner, mint)
    print("Token Account:", token_account)

    if not token_account:
        print("Failed to retrieve or create token account.")
        return False

    virtual_sol_reserves = coin_data['virtual_sol_reserves']
    virtual_token_reserves = coin_data['virtual_token_reserves']
    sol_in_lamports = sol_in * LAMPORTS_PER_SOL
    amount = int(sol_in_lamports * virtual_token_reserves / virtual_sol_reserves)
    print("Calculated Amount:", amount)

    buildTxn(mint, amount, slippage, token_account, sol_in, token_account_instructions, buy=True)

    time.sleep(2)

    return True
pump_fun_buy("4dCfQ7pSPihS4LnZjYgXznjJ3yEEwQbKQTbCfin9pump")



from .constants import *
def buy(mint_str: str, sol_in: float = 0.001, slippage: int = 25) -> bool:
   
        # Get coin data
        coin_data = get_pump_fun_data(mint_str)
        if not coin_data:
            print("Failed to retrieve coin data...")
            return False
        payer_keypair = get_payer_keypair()
        owner = payer_keypair.pubkey()
        mint = Pubkey.from_string(mint_str)
        # Attempt to retrieve token account; if not, prepare to create it
        token_account = get_associated_token_address(owner, mint)
        account_data = get_token_accounts_by_owner(owner,mint)
        token_account_instructions = None

        if not account_data.get('value'):
            # If token account does not exist, create the associated token account
            print("Creating new associated token account...")
            token_account_instructions = create_associated_token_account(owner, owner, mint)

        # Calculate amount of tokens to buy
        virtual_sol_reserves = coin_data['virtual_sol_reserves']
        virtual_token_reserves = coin_data['virtual_token_reserves']
        sol_in_lamports = sol_in * LAMPORTS_PER_SOL
        amount = int(sol_in_lamports * virtual_token_reserves / virtual_sol_reserves)

        # Calculate max SOL cost with slippage
        max_sol_cost = int(sol_in * (1 + slippage / 100) * LAMPORTS_PER_SOL)
        print("Max SOL Cost:", max_sol_cost / LAMPORTS_PER_SOL)

        # Define account keys for the transaction
        keys = [
            AccountMeta(pubkey=GLOBAL, is_signer=False, is_writable=False),
            AccountMeta(pubkey=FEE_RECIPIENT, is_signer=False, is_writable=True),
            AccountMeta(pubkey=Pubkey.from_string(coin_data['mint']), is_signer=False, is_writable=False),
            AccountMeta(pubkey=Pubkey.from_string(coin_data['bonding_curve']), is_signer=False, is_writable=True),
            AccountMeta(pubkey=Pubkey.from_string(coin_data['associated_bonding_curve']), is_signer=False, is_writable=True),
            AccountMeta(pubkey=token_account, is_signer=False, is_writable=True),
            AccountMeta(pubkey=owner, is_signer=True, is_writable=True),
            AccountMeta(pubkey=SYSTEM_PROGRAM, is_signer=False, is_writable=False),
            AccountMeta(pubkey=TOKEN_PROGRAM, is_signer=False, is_writable=False),
            AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
            AccountMeta(pubkey=EVENT_AUTHORITY, is_signer=False, is_writable=False),
            AccountMeta(pubkey=PUMP_FUN_PROGRAM_ID, is_signer=False, is_writable=False)
        ]

        # Construct the swap instruction
        data = bytearray()
        data.extend(bytes.fromhex("66063d1201daebea"))  # Replace with actual instruction identifier
        data.extend(struct.pack('<Q', amount))
        data.extend(struct.pack('<Q', max_sol_cost))
        swap_instruction = Instruction(PUMP_FUN_PROGRAM_ID, bytes(data), keys)

        # Construct and sign transaction
        txn = Transaction(recent_blockhash=get_block_hash(), fee_payer=owner)
        txn.add(set_compute_unit_price(UNIT_PRICE))
        txn.add(set_compute_unit_limit(UNIT_BUDGET))
        if token_account_instructions:
            txn.add(token_account_instructions)
        txn.add(swap_instruction)

        # Sign and send transaction
        txn.sign(payer_keypair)
        txn_sig = sendTransaction(txn, payer_keypair, opts=TxOpts(skip_preflight=True))
        print("Transaction Signature:", txn_sig)
        return txn_sig



from utils import *
GLOBAL = get_pubKey("4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf")
FEE_RECIPIENT = get_pubKey("CebN5WGQ4jvEPvsVU4EoHEpgzq1VV7AbicfhtW4xC9iM")
SYSTEM_PROGRAM = get_pubKey("11111111111111111111111111111111")
TOKEN_PROGRAM = get_pubKey("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")
ASSOC_TOKEN_ACC_PROG = get_pubKey("ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL")
RENT = get_pubKey("SysvarRent111111111111111111111111111111111")
EVENT_AUTHORITY = get_pubKey("Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1")
PUMP_FUN_PROGRAM = get_pubKey("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P")

GLOBAL = get_pubKey("4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf")
FEE_RECIPIENT = get_pubKey("CebN5WGQ4jvEPvsVU4EoHEpgzq1VV7AbicfhtW4xC9iM")
SYSTEM_PROGRAM = get_pubKey("11111111111111111111111111111111")
TOKEN_PROGRAM = get_pubKey("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")
ASSOC_TOKEN_ACC_PROG = get_pubKey("ATokenGPuRNW6R3UB3MBP2ZsLvRkYJLpUzB6bZyP4jE")
RENT = get_pubKey("SysvarRent111111111111111111111111111111111")
EVENT_AUTHORITY = get_pubKey("Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1")
PUMP_FUN_PROGRAM = get_pubKey("6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P")
LAMPORTS_PER_SOL = 1_000_000_000
UNIT_PRICE =  1_000_000
UNIT_BUDGET =  100_000



from solana.rpc.api import Client
from solders.keypair import Keypair #type: ignore
from abstract_security import *
from solders.pubkey import Pubkey
def load_from_private_key(env_key=None):
    envKey = env_key or 'AMM_P'
    env_value = get_env_value(key=envKey)
    if env_value:
        return Keypair.from_base58_string(env_value)
def check_existing_token_account(owner: Pubkey, mint: Pubkey):
    try:
        account_data = get_account_by_owner(str(owner), str(mint))
        if account_data:
            token_account = account_data['pubkey']
            print(f"Existing token account found: {token_account}")
            return token_account, None
        else:
            print("No existing token account found. Creating a new one...")
            token_account = get_associated_token_address(owner, mint)
            token_account_instructions = create_associated_token_account(owner, owner, mint)
            return token_account, token_account_instructions
    except Exception as e:
        print(f"Error checking or creating token account: {e}")
        return None, None
def load_keypair_from_file(filename):
    curr = os.path.join(sys.path[0], 'data',  filename)
    with open(curr, 'r') as file:
        secret = json.load(file)
        secret_key = bytes(secret)
        print(base58.b58encode(secret_key))
        return Keypair.from_bytete_key()
def get_payer_keypair(fileName=None,envKey=None):

    payerKeypair = None
    if fileName:
        payerKeypair = load_keypair_from_file(filename)
    if payerKeypair is None:
        payerKeypair = load_from_private_key(env_key=envKey)
    return payerKeypair
payer_keypair =  load_from_private_key()



import base64
from spl.token.instructions import get_associated_token_address
from construct import Padding, Struct, Int64ul, Flag
from constants import PUMP_FUN_PROGRAM,get_pubKey,Pubkey
from isolated_rpcs import getAccountInfo
# Define the struct format for virtual reserves
bonding_curve_struct = Struct(
    Padding(8),
    "virtual_token_reserves" / Int64ul,
    "virtual_sol_reserves" / Int64ul,
    "real_token_reserves" / Int64ul,
    "real_sol_reserves" / Int64ul,
    "token_total_supply" / Int64ul,
    "complete" / Flag
)

    
def get_virtual_reserves(bonding_curve: Pubkey):
    # Convert Pubkey to string for JSON compatibility
    bonding_curve_str = str(bonding_curve)
    
    account_info = getAccountInfo(bonding_curve_str)
    
    # Check if account info retrieval was successful
    if not account_info or 'data' not in account_info:
        print("Failed to retrieve account info.")
        return None

    try:
        # Decode base64 data and parse it using the struct format
        data_base64 = account_info['data'][0]
        data = base64.b64decode(data_base64)
        parsed_data = bonding_curve_struct.parse(data)
    except Exception as e:
        print(f"Error decoding and parsing data: {e}")
        return None

    return parsed_data

def derive_bonding_curve(mint):
    # Find bonding curve address based on mint
    return Pubkey.find_program_address(["bonding-curve".encode(), bytes(get_pubKey(mint))], PUMP_FUN_PROGRAM)

def derive_bonding_curve_accounts(mint_str: str):
    mint = get_pubKey(mint_str)
    bonding_curve, _ = derive_bonding_curve(mint)
    associated_bonding_curve = get_associated_token_address(bonding_curve, mint)
    return {'bonding_curve': bonding_curve, "associated_bonding_curve": associated_bonding_curve}

def get_pump_fun_data(mint_str: str):
    # Derive bonding curve and associated account
    bonding_curve_data = derive_bonding_curve_accounts(mint_str)
    bonding_curve = bonding_curve_data['bonding_curve']
    associated_bonding_curve = bonding_curve_data['associated_bonding_curve']
    
    # Retrieve virtual reserves
    virtual_reserves = get_virtual_reserves(bonding_curve)
    
    if virtual_reserves is None:
        print("Failed to retrieve virtual reserves data.")
        return None

    # Extract reserves data into dictionary
    txn_data = {
        "mint": mint_str,
        "bonding_curve": str(bonding_curve),
        "associated_bonding_curve": str(associated_bonding_curve),
        "virtual_token_reserves": virtual_reserves.virtual_token_reserves,
        "virtual_sol_reserves": virtual_reserves.virtual_sol_reserves,
        "token_total_supply": virtual_reserves.token_total_supply,
        "complete": virtual_reserves.complete
    }
    
    return txn_data
